( cd usr/man/preformat/cat1 ; rm -rf XF86_8514.1x.gz )
( cd usr/man/preformat/cat1 ; ln -sf XF86_Accel.1x.gz XF86_8514.1x.gz )
( cd usr/man/preformat/cat1 ; rm -rf XF86_AGX.1x.gz )
( cd usr/man/preformat/cat1 ; ln -sf XF86_Accel.1x.gz XF86_AGX.1x.gz )
( cd usr/man/preformat/cat1 ; rm -rf XF86_Mach32.1x.gz )
( cd usr/man/preformat/cat1 ; ln -sf XF86_Accel.1x.gz XF86_Mach32.1x.gz )
( cd usr/man/preformat/cat1 ; rm -rf XF86_Mach8.1x.gz )
( cd usr/man/preformat/cat1 ; ln -sf XF86_Accel.1x.gz XF86_Mach8.1x.gz )
( cd usr/man/preformat/cat1 ; rm -rf XF86_P9000.1x.gz )
( cd usr/man/preformat/cat1 ; ln -sf XF86_Accel.1x.gz XF86_P9000.1x.gz )
( cd usr/man/preformat/cat1 ; rm -rf XF86_S3.1x.gz )
( cd usr/man/preformat/cat1 ; ln -sf XF86_Accel.1x.gz XF86_S3.1x.gz )
( cd usr/man/preformat/cat1 ; rm -rf XF86_W32.1x.gz )
( cd usr/man/preformat/cat1 ; ln -sf XF86_Accel.1x.gz XF86_W32.1x.gz )
